#!/usr/bin/env python3
from .appdirs import user_config_dir as appdirs
import os
import sys
import pathlib
from shlex import split
import csv


class UserSettings:
    """Class for managing user config files"""

    # User files containing their location on the system
    conf_dir = None
    conf_file = None

    # Command line args passed by user
    # This is an argparse object
    cmd_args = None

    # User's settings as set in their conf and filters files
    prompt_cmd = None

    def __init__(self, program, cmd_line_args=None):
        self.add_cmd_args(cmd_line_args=cmd_line_args)
        self.set_files(program=program)
        self.validate_config()
        self.read_config()

    @classmethod
    def add_cmd_args(cls, cmd_line_args=None):
        cls.cmd_args = cmd_line_args

    @classmethod
    def set_files(cls, program=None):
        """Return user config directory and config_files dict"""

        cls.conf_dir = appdirs(program)
        config_files = {
            "conf_file": f"{program}.conf",
        }

        for key, value in config_files.items():
            value = os.path.join(cls.conf_dir, value)
            setattr(cls, key, value)

    @classmethod
    def read_config(cls):
        """Read user config files and update user_args"""

        with open(cls.conf_file) as conf:
            reader = csv.reader(
                conf, delimiter="=", escapechar="\\", quoting=csv.QUOTE_NONE
            )

            config_settings = ["prompt_cmd"]

            for row in reader:
                if len(row) > 2:
                    raise csv.Error(f"Too many fields on row: {row}")

                setting_name = row[0].strip().lower()
                setting_value = row[1].strip()

                if not setting_name in config_settings:
                    continue

                setattr(cls, setting_name, setting_value)

    @classmethod
    def validate_config(cls):
        """Test if config files exist and create them if needed"""

        dir = pathlib.Path(cls.conf_dir)
        if not dir.is_dir():
            dir.mkdir(parents=True, exist_ok=True)

        files = [cls.conf_file]

        for file in files:
            f = pathlib.Path(file)

            if not f.is_file():
                f.touch()



#!/usr/bin/env python3
from .dmenu import show as dmenu_show
from .dmenu import DmenuCommandError, DmenuUsageError, DmenuError
from .fzf import FzfPrompt


class InvalidCmdPrompt(Exception):
    """Exception raised when user has invalid command prompt"""

    def __init__(self, error, message="ERROR: prompt_cmd not recognized"):
        self.error = error
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} "{self.error}"'


class InputError(Exception):
    """Exception fzf or dmenu prompt fails"""

    def __init__(self, error, message="ERROR: Could not get user input", prefix=""):
        self.error = error
        self.prefix = prefix
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.prefix}{self.message}{self.error}"


def dmenu_prompt(
    items,
    command="dmenu",
    command_args=[],
    case_insensitive=None,
    lines=None,
    monitor=None,
    prompt=None,
):

    choice = None
    try:
        choice = dmenu_show(
            items=items,
            command=command,
            command_args=command_args,
            case_insensitive=case_insensitive,
            lines=lines,
            monitor=monitor,
            prompt=prompt,
        )

    except (DmenuCommandError, DmenuUsageError, DmenuError) as err:
        raise InputError(err)

    except KeyboardInterrupt:
        raise InputError("User interruption...stopping", message="")

    return choice


def fzf_auto(options, prompt):
    choice = None
    try:
        print(f"Auto title: {options[0]}")
        choice = input(prompt)
        if choice == "":
            choice = options[0]

    except KeyboardInterrupt:
        raise InputError("User interruption...stopping", message="", prefix="\n")

    return choice


def fzf_prompt(options, prompt):
    choice = None
    try:
        fzf = FzfPrompt()
        choice = fzf.prompt(options, f'--prompt="{prompt}"')

    except SystemError as err:
        raise InputError(err)

    return choice


def user_choice(
    options,
    prompt_cmd,
    prompt,
    other_message=None,
    other_position="last",
    auto_title=False,
):
    """
    Give user a prompt to choose from a list of options.
    Uses dmenu, fzf, or rofi
    """
    if prompt_cmd == "dmenu":
        choice = dmenu_prompt(
            items=options,
            command=prompt_cmd,
            prompt=prompt,
            lines=20,
            case_insensitive=True,
        )

    elif prompt_cmd == "rofi":
        choice = dmenu_prompt(
            items=options, command=prompt_cmd, prompt=prompt, command_args=["-dmenu"]
        )

    elif prompt_cmd == "fzf":
        if auto_title:
            choice = fzf_auto(options, prompt)

        else:
            choice = fzf_prompt(options, prompt)
            if len(choice) == 1:
                choice = choice[0]

    else:
        raise InvalidCmdPrompt(prompt_cmd)

    return choice


def zathura_prompt(pid_dict, prompt_cmd):
    """
    Prompt user to select their desired instance
    """
    # Create a list of options made up of the second value of
    # z.pid_dict. This looks like "123456: /path/to/file.pdf"
    options = []
    for val in pid_dict.values():
        options.append(val[1])

    # Ask user about what instance to use
    prompt = "Use: "
    try:
        choice = user_choice(
            options=options,
            prompt_cmd=prompt_cmd,
            prompt=prompt,
        )
    except (InvalidCmdPrompt, InputError, KeyboardInterrupt) as err:
        print(err, sys.stderr)
        return None

    # Search for the chosen instance and return it's pid
    for key, val in pid_dict.items():
        if val[1] == choice:
            return key

    return None



#!/usr/bin/env python3
import os
import sys
import psutil
import dbus
import subprocess
import pathlib
import time
import tempfile
from importlib.util import find_spec
from .zsocket import SocketListener


if find_spec("ewmh"):
    from ewmh import EWMH

    has_ewmh = True
else:
    has_ewmh = False


class Zathura:
    """
    Class to interact with zathura
    methods:
        self._mk_pid_file()     # Used internally to create files
                                # where znp -w $TMP_FILE $FILE
                                # will wite the value of $FILE

        self.ewmh_get_pid()     # Used to get the pid of the
                                # active window if running X11
                                # and ewmh is installed

        self.execute()          # Takes a pid and a command to run
                                # from the given instance

        self.open()             # Takes a pid and a file_path to be
                                # openend in the given instance

        self.pid_search()       # Runs when class is created
                                # adds a list of zpids to self.pids

        self.remove_tmp_files() # Removes all files in self.files_to_remove

        self.zathura_search()   # Takes a pid and optionaly a search file
                                # and tries to find a zathura instance
                                # that has that file currently open
    attributes:
        self.pids list          # Is a list of pids that belong to
                                # running zathura instances

        self.has_ewmh bool      # True if user has ewmh installed

        self.epid int           # Is a pid value stored after running
                                # self.ewmh_get_pid()

        self.pid_files dict     # Is a dictionary of pids keys with
                                # file values

        self.pid_dict dict      # Is a dictionary of pid keys with
                                # a list value where the first value in
                                # the list is the currently open file
                                # and the second value is the pid tied
                                # to the file. Multiple pids may have the
                                # same file open and this ensures the desired
                                # pid is chosen

        self.files_to_remove    # Is a list of files that need to be removed
                                # if self.z_search is run
    """

    def __init__(self):
        self.has_ewmh = has_ewmh
        self.pid_search()

    def _mk_pid_file(self, file_path):
        """Create the given file if it does not exist"""
        if os.path.isfile(file_path):
            return
        fp = pathlib.Path(file_path).touch(mode=0o600)
        return

    def ewmh_get_pid(self):
        """
        Use ewmh to get the pid of the focused window.
        This should be zathura assuming znp was run from
        zathura and user is using X11
        """
        if not self.has_ewmh:
            return False
        ewmh = EWMH()
        win = ewmh.getActiveWindow()
        pid = ewmh.getWmPid(win)
        if pid in self.pids:
            self.epid = pid
            return True
        self.epid = None
        return False

    def execute(self, pid, command):
        """
        Execute a zathura command. Requires the given zpid to send
        the command to.
        """
        zathuraName = "org.pwmt.zathura.PID-" + str(pid)
        zathuraObject = dbus.SessionBus().get_object(zathuraName, "/org/pwmt/zathura")
        zathuraDBus = dbus.Interface(zathuraObject, "org.pwmt.zathura")
        zathuraDBus.ExecuteCommand(command)
        return 0

    def open(self, file_path, pid=None):
        """
        Open file in zathura optionally open the next file in
        the directory of the given file
        """
        # If zathura isn't open we can open it
        if not pid:
            zathuraPopen = subprocess.Popen(["zathura"])
            pid = zathuraPopen.pid
            time.sleep(0.5)
        # Open the given file
        zathuraName = "org.pwmt.zathura.PID-" + str(pid)
        zathuraObject = dbus.SessionBus().get_object(zathuraName, "/org/pwmt/zathura")
        zathuraDBus = dbus.Interface(zathuraObject, "org.pwmt.zathura")
        zathuraDBus.OpenDocument(file_path, "", 0)
        return 0

    def pid_search(self, name="zathura"):
        """
        Search for zathura pid(s). Used when:
            1. User is using Wayland
            2. ewmh fails for some reason
            3. If the user did not provide a PID at runtime
        """
        pids = []
        for proc in psutil.process_iter():
            try:
                p_name = proc.name()
            except psutil.NoSuchProcess:
                continue
            if p_name == name:
                pids.append(proc.pid)
        self.pids = pids
        self.pid_count = len(pids)
        return None

    def query(self, search_file):
        """
        Query zathura instances for the instance with search_file open.
        If found prints the pid, else nothing and exits with status 1.
        """
        if self.pid_count == 0:
            return 1
        elif self.pid_count == 1:
            print(self.pids[0])
            return 0
        elif self.zathura_search(self.pids, search_file=search_file):
            print(self.search_pid)
            return 0

        return 1

    def zathura_search(self, pids, search_file=None, prog="znp"):
        """
        Search for an instance of zathura using bad means.
        This creates a dictionary of zathura instances if multiple exist:
            >>> zathura = Zathura()
            >>> print(zathura.pid_dict)
            >>> '{ "123456": ["/path/to/file.pdf", "123456: /path/to/file.pdf"]
                   "154684": ["/path/to/file.epub", "154684: /path/to/file.epub"] }'
        """
        # Create temp dir for placing socket
        tdir = tempfile.gettempdir()
        tdir = os.path.join(tdir, prog)
        server_address = os.path.join(tdir, "znp_socket")
        if not os.path.isdir(tdir):
            os.mkdir(tdir, mode=0o700)
        # Start SocketListener and expect as many messages as pids
        pid_count = len(pids)
        server = SocketListener(
            server_address=server_address,
            expected_messages=pid_count,
        )
        # Give server some time to start
        time.sleep(0.05)
        # Run clients to find current files in each pid
        for pid in pids:
            self.execute(pid, f"exec znp -P {pid} -c {server_address} $FILE")
        # Wait for server to get all client messages
        while server.recieved < pid_count:
            pass
        # Server must have recieved all messages, store it's data
        self.pid_dict = server.pid_dict
        self.search_pid = None
        # If not searching for a specific file return True
        if not search_file:
            return True
        # Search the pid_dict for the desired file
        for key, value in server.pid_dict.items():
            if value[0] == search_file:
                # Search_file found set search_pid and return True
                self.search_pid = key
                return True
        # Desired instance wasn't found return False
        return False



#!/usr/bin/env python3
import argparse


def get_opts(prog_name="znp"):
    parser = argparse.ArgumentParser(
        prog=prog_name,
        description="""
        Add a file to a given instance of zathura
        """,
        allow_abbrev=False,
    )
    parser.add_argument(
        "-c",
        "--client",
        metavar="SERVER_ADDRESS",
        action="store",
        default=None,
        help="""
        Send a message to the socket at SERVER_ADDRESS. Message has a seven byte header
        which defaults to 0000000 if PID is not given. FILE is the message that will be
        sent and need not be an actual message. Used internally when searching for active
        instances and their open files and is executed like so:
        `znp -P 1234567 -c "/tmp/znp/znp_socket" "/path/to/current.pdf"` from the instance
        that belongs to pid 1234567.
        """,
    )
    parser.add_argument(
        "-n",
        "--next",
        action="store_true",
        help="""
        Go to the next file in the directory of the given file
        """,
    )
    parser.add_argument(
        "-p",
        "--prev",
        action="store_true",
        help="""
        Go to the prev file in the directory of the given file
        """,
    )
    parser.add_argument(
        "-P",
        "--pid",
        metavar="PID",
        action="store",
        default=None,
        help="""
        PID of the zathura instance to use. Default is the
        active window detected via ewmh
        """,
    )
    parser.add_argument(
        "-q",
        "--query",
        action="store_true",
        help="""
        Query zathura instances for the instance with FILE open.
        If found prints the pid, else prints nothing and exits with status 1.
        """,
    )
    parser.add_argument(
        "-s",
        "--search",
        action="store_true",
        help="""
        Print a list of Zathura pids
        """,
    )
    parser.add_argument(
        "file",
        metavar="FILE",
        help="""
        The file to open, or the file that is currently open.
        """,
    )
    args = parser.parse_args()
    return args
